﻿Imports System.Collections.Generic

''' <summary>
''' Direct3Dを扱うライブラリ
''' 制作：
''' </summary>
''' <remarks></remarks>
Namespace DirectGameFramework

    ''' <summary>
    ''' メインループ、描画を扱うクラス
    ''' </summary>
    ''' <remarks></remarks>
    Public Class Main_Render
        Inherits Form
        Implements IDisposable




        ''' <summary>
        ''' モデルクラス
        ''' </summary>
        ''' <remarks></remarks>
        Public Class Model
            Public Mesh As Mesh
            Public Material() As Material
            Public Texture() As Texture
            Public Pos As Vector3
            Public Rotation As Vector3
            Public Scaling As Vector3
            Public Matrix As Matrix
            Public BBMin As Vector3
            Public BBMax As Vector3
            Public Visible As Boolean
            Public Free As Integer
            Public Free2 As Integer
        End Class

#Region "宣言、プロパティ"

        ''' <summary>
        ''' 描画デバイス
        ''' </summary>
        Public Device As Direct3D.Device

        Public PM As PresentParameters

        ''' <summary>
        ''' 汎用スプライト
        ''' </summary>
        Public Sprite As Sprite


        ''' <summary>
        ''' 登録されたフォント
        ''' </summary>
        Public Fonts As New List(Of Direct3D.Font)

        ''' <summary>
        ''' 登録されたテクスチャー
        ''' </summary>
        Public Textures As New List(Of Texture)

        Delegate Sub Get_Sub()
        Dim SetDatas As Get_Sub
        Dim Render As Get_Sub

        Delegate Sub FormClosingEvent_Sub(ByVal e As Windows.Forms.FormClosingEventArgs)
        Delegate Sub KeyEvent_Sub(ByVal e As Windows.Forms.KeyEventArgs)
        Delegate Sub ClickEvent_Sub(ByVal e As Windows.Forms.MouseEventArgs)

        Dim ClosingEvent As FormClosingEvent_Sub
        Dim KeyEvent As KeyEvent_Sub
        Dim ClickEvent As ClickEvent_Sub

        Dim FPSTimer As Timer = New Timer() 'FPS計測用
        Dim FPSTimeri As Integer
        Dim FPSi As Integer


        ''' <summary>
        ''' 関数をセットすると1フレーム頭ごとに実行される
        ''' </summary>
        WriteOnly Property SetData_Sub() As Get_Sub
            Set(ByVal value As Get_Sub)
                SetDatas = value
            End Set
        End Property

        ''' <summary>
        ''' 関数をセットすると1フレームごとに実行される
        ''' </summary>
        WriteOnly Property Render_Sub() As Get_Sub
            Set(ByVal value As Get_Sub)
                Render = value
            End Set
        End Property

        ''' <summary>
        ''' 関数をセットするとclosingイベント時実行される
        ''' </summary>
        WriteOnly Property Closing_Sub() As FormClosingEvent_Sub
            Set(ByVal value As FormClosingEvent_Sub)
                ClosingEvent = value
            End Set
        End Property

        ''' <summary>
        ''' 関数をセットするとKey_Downイベント時実行される
        ''' </summary>
        WriteOnly Property Key_Sub() As KeyEvent_Sub
            Set(ByVal value As KeyEvent_Sub)
                KeyEvent = value
            End Set
        End Property

        ''' <summary>
        ''' 関数をセットするとClickイベント時実行される
        ''' </summary>
        WriteOnly Property Click_Sub() As ClickEvent_Sub
            Set(ByVal value As ClickEvent_Sub)
                ClickEvent = value
            End Set
        End Property

        ''' <summary>
        ''' 現在のフレームレートを取得する
        ''' </summary>
        ReadOnly Property FPS() As Integer
            Get
                Return FPSi
            End Get
        End Property
#End Region

        ''' <summary>
        ''' 800x600の画面サイズでクラスを初期化する
        ''' </summary>
        Public Sub New()
            Me.New(800, 600, "Direct3D")
        End Sub

        ''' <summary>
        ''' 指定した画面サイズでクラスを初期化する
        ''' </summary>
        ''' <param name="Width">画面サイズwidth</param>
        ''' <param name="Height">画面サイズheight</param>
        ''' <param name="Text">ウィンドウタイトル</param>
        Public Sub New(ByVal Width As Integer, ByVal Height As Integer, ByVal Text As String, Optional ByVal BGI As String = "")

            Me.ClientSize = New Size(Width, Height)
            Me.FormBorderStyle = Windows.Forms.FormBorderStyle.FixedSingle
            Me.MaximizeBox = False
            Me.Text = Text
            If Not BGI = "" Then Me.BackgroundImage = Image.FromFile(BGI)

            PM = New PresentParameters
            With PM
                .Windowed = True
                .BackBufferWidth = Width
                .BackBufferHeight = Height
                .BackBufferFormat = Format.X8R8G8B8
                .SwapEffect = SwapEffect.Discard
                .EnableAutoDepthStencil = True
                .AutoDepthStencilFormat = DepthFormat.D16
            End With

            Try
                Device = New Direct3D.Device(0, Direct3D.DeviceType.Hardware, Me, CreateFlags.HardwareVertexProcessing, PM)
            Catch
                Try
                    Device = New Direct3D.Device(0, Direct3D.DeviceType.Hardware, Me, CreateFlags.SoftwareVertexProcessing, PM)
                Catch
                    MsgBox("初期化に失敗しました")
                    Application.Exit()
                End Try
            End Try

            Device.RenderState.ZBufferEnable = True
            Device.RenderState.FogEnable = True

            Me.Show()

            With Device.Lights(0)
                .Type = Microsoft.DirectX.Direct3D.LightType.Directional
                .Diffuse = Drawing.Color.White
                .Direction = New Vector3(0, -1, 0)
                .Enabled = True
                .Update()
            End With
            Device.RenderState.Ambient = Drawing.Color.White

            With Device.Transform
                .View = Matrix.LookAtLH(New Vector3(0, 0, -5), New Vector3(0, 0, 0), New Vector3(0, 1, 0))
                .Projection = Matrix.PerspectiveFovLH(Math.PI / 4, Me.Width / Me.Height, 0.1, 100)
            End With

            AddHandler FPSTimer.Tick, AddressOf Timer_Tick
            FPSTimer.Interval = 1000
            FPSTimer.Enabled = True

            Sprite = New Sprite(Device)

        End Sub

        ''' <summary>
        ''' デバイスをリセットする
        ''' </summary>
        ''' <param name="prm">デバイスに与えるPresentParameters</param>
        Public Sub DeviceReset(ByVal prm As PresentParameters)
            Device.Reset(prm)
        End Sub

        ''' <summary>
        ''' メインループ（の起動）
        ''' </summary>
        Public Sub MainLoop()

            While Me.Created

                If Not SetDatas Is Nothing Then SetDatas()

                Renders()


                Application.DoEvents()

                FPSTimeri += 1



            End While

        End Sub

        ''' <summary>
        ''' 描画
        ''' </summary>
        Sub Renders()
            If Not Render Is Nothing Then
                Render()
                Exit Sub
            End If

            Device.Clear(ClearFlags.Target Or ClearFlags.ZBuffer, Color.Gray, 1, 0)

        End Sub

#Region "モデルリスト操作"


        ''' <summary>
        ''' Xファイルからののモデル新規追加
        ''' </summary>
        ''' <param name="List">操作するモデルリスト</param>
        ''' <param name="Name">Xファイルのパス</param>
        ''' <param name="PosX">位置X</param>
        ''' <param name="PosY">位置Y</param>
        ''' <param name="PosZ">位置Z</param>
        ''' <param name="RotX">回転X</param>
        ''' <param name="RotY">回転Y</param>
        ''' <param name="RotZ">回転Z</param>
        ''' <param name="ScaX">拡縮X</param>
        ''' <param name="ScaY">拡縮Y</param>
        ''' <param name="ScaZ">拡縮Z</param>
        ''' <returns>格納されたインデックス</returns>
        Public Function ModelToXFile(ByRef List As List(Of Model), ByVal Name As String, Optional ByVal Collide As Boolean = False, Optional ByVal Visible As Boolean = True, Optional ByVal PosX As Double = 0, Optional ByVal PosY As Double = 0, Optional ByVal PosZ As Double = 0, Optional ByVal RotX As Double = 0, Optional ByVal RotY As Double = 0, Optional ByVal RotZ As Double = 0, Optional ByVal ScaX As Double = 1, Optional ByVal ScaY As Double = 1, Optional ByVal ScaZ As Double = 1)

            Dim Mat() As ExtendedMaterial
            Dim Model As New Model

            Model.Mesh = Mesh.FromFile(Name, MeshFlags.SystemMemory, Device, Mat)
            'テクスチャ
            If Model.Texture Is Nothing Then

                Model.Material = New Material(Mat.Length) {}
                Model.Texture = New Texture(Mat.Length) {}

                For i As Integer = 0 To Mat.Length - 1

                    Model.Material(i) = Mat(i).Material3D
                    Model.Material(i).Ambient = Model.Material(i).Diffuse

                    If Mat(i).TextureFilename = Nothing Then Continue For
                    Try
                        Model.Texture(i) = TextureLoader.FromFile(Device, Mat(i).TextureFilename)
                    Catch ex As Exception
                        Model.Texture(i) = TextureLoader.FromFile(Device, IO.Path.GetDirectoryName(Name) & "\" & Mat(i).TextureFilename)

                    End Try

                Next
            End If

            Model.Pos = New Vector3(PosX, PosY, PosZ)
            Model.Rotation = New Vector3(RotX, RotY, RotZ)
            Model.Scaling = New Vector3(ScaX, ScaY, ScaZ)

            Model.Matrix = New Matrix
            Model.Matrix = Matrix.Identity
            Model.Visible = Visible

            If Collide Then Get_BoundingBox(Model, False)

            List.Add(Model)
            Return List.Count - 1

        End Function

        Public Sub ReLoadModelToXFile(ByRef List As List(Of Model), ByVal Index As Double, ByVal Name As String, Optional ByVal Collide As Boolean = False, Optional ByVal Visible As Boolean = True, Optional ByVal PosX As Double = 0, Optional ByVal PosY As Double = 0, Optional ByVal PosZ As Double = 0, Optional ByVal RotX As Double = 0, Optional ByVal RotY As Double = 0, Optional ByVal RotZ As Double = 0, Optional ByVal ScaX As Double = 1, Optional ByVal ScaY As Double = 1, Optional ByVal ScaZ As Double = 1)

            Dim Mat() As ExtendedMaterial
            Dim Model As New Model

            Model.Mesh = Mesh.FromFile(Name, MeshFlags.SystemMemory, Device, Mat)
            'テクスチャ
            If Model.Texture Is Nothing Then

                Model.Material = New Material(Mat.Length) {}
                Model.Texture = New Texture(Mat.Length) {}

                For i As Integer = 0 To Mat.Length - 1

                    Model.Material(i) = Mat(i).Material3D
                    Model.Material(i).Ambient = Model.Material(i).Diffuse

                    If Mat(i).TextureFilename = Nothing Then Continue For
                    Try
                        Model.Texture(i) = TextureLoader.FromFile(Device, Mat(i).TextureFilename)
                    Catch ex As Exception
                        Model.Texture(i) = TextureLoader.FromFile(Device, IO.Path.GetDirectoryName(Name) & "\" & Mat(i).TextureFilename)

                    End Try

                Next
            End If

            Model.Pos = New Vector3(PosX, PosY, PosZ)
            Model.Rotation = New Vector3(RotX, RotY, RotZ)
            Model.Scaling = New Vector3(ScaX, ScaY, ScaZ)

            Model.Matrix = New Matrix
            Model.Matrix = Matrix.Identity

            Model.Visible = Visible

            If Collide Then Get_BoundingBox(Model, False)

            List(Index) = Model

        End Sub


        ''' <summary>
        ''' 登録されたモデルを描画
        ''' </summary>
        ''' <param name="List">操作するモデルリスト</param>
        ''' <param name="index">描画するXファイルのインデックス</param>
        Public Sub RenderModel(ByRef List As List(Of Model), ByVal index As Integer)
            If List(index).Visible = False Then Exit Sub
            Device.Transform.World = List(index).Matrix
            Device.Transform.World *= Matrix.Scaling(List(index).Scaling)
            Device.Transform.World *= Matrix.RotationX(List(index).Rotation.X)
            Device.Transform.World *= Matrix.RotationY(List(index).Rotation.Y)
            Device.Transform.World *= Matrix.RotationZ(List(index).Rotation.Z)
            Device.Transform.World *= Matrix.Translation(List(index).Pos)


            For i As Integer = 0 To List(index).Material.Length - 1
                Device.Material = List(index).Material(i)
                Device.SetTexture(0, List(index).Texture(i))
                List(index).Mesh.DrawSubset(i)
            Next

        End Sub

        ''' <summary>
        ''' 登録されたXファイルをすべて描画
        ''' <param name="List">操作するモデルリスト</param>
        ''' </summary>
        Public Sub RenderModels(ByRef List As List(Of Model))

            For Each X As Model In List
                Threading.Thread.Sleep(1)
                If X.Visible = False Then Continue For
                Device.Transform.World = X.Matrix
                Device.Transform.World *= Matrix.Scaling(X.Scaling)
                Device.Transform.World *= Matrix.RotationX(X.Rotation.X)
                Device.Transform.World *= Matrix.RotationY(X.Rotation.Y)
                Device.Transform.World *= Matrix.RotationZ(X.Rotation.Z)
                Device.Transform.World *= Matrix.Translation(X.Pos)

                For i As Integer = 0 To X.Material.Length - 1
                    Device.Material = X.Material(i)
                    Device.SetTexture(0, X.Texture(i))
                    X.Mesh.DrawSubset(i)
                Next
            Next
        End Sub

#End Region


        ''' <summary>
        ''' モデル同士の当たり判定をとる
        ''' </summary>
        ''' <param name="objA">モデルA</param>
        ''' <param name="objB">モデルB</param>
        Public Function Get_Collide(ByVal objA As Model, ByVal objB As Model) As Boolean

            If objA.Visible = False Or objB.Visible = False Then Exit Function

            Dim MaxA, MinA, MaxB, MinB As Vector3


            MaxA = objA.BBMax + objA.Pos
            MinA = objA.BBMin + objA.Pos
            MaxB = objB.BBMax + objB.Pos
            MinB = objB.BBMin + objB.Pos


            If (MinA.X < MaxB.X) And (MinA.Y < MaxB.Y) And (MinA.Z < MaxB.Z) _
            And (MinB.X < MaxA.X) And (MinB.Y < MaxA.Y) And (MinB.Z < MaxA.Z) Then
                Return True
            End If

            Return False

        End Function


        ''' <summary>
        ''' モデルにBoundingBoxを登録
        ''' </summary>
        ''' <param name="obj">モデル</param>
        ''' <param name="Trans">変形後の計算をするか</param>
        Public Sub Get_BoundingBox(ByVal obj As Model, ByVal Trans As Boolean)
            Dim Min, Max As Vector3
            Dim vertbuf As VertexBuffer

            If Trans Then
                vertbuf = TransMesh(obj)
            Else
                vertbuf = obj.Mesh.VertexBuffer
            End If

            Dim verts As GraphicsStream = vertbuf.Lock(0, 0, LockFlags.None)
            Geometry.ComputeBoundingBox(verts, obj.Mesh.NumberVertices, vertbuf.Description.VertexFormat, Min, Max)

            verts.Close()
            vertbuf.Unlock()

            obj.BBMin = Min
            obj.BBMax = Max

        End Sub

        ''' <summary>
        ''' モデルを変形しそのvertexbufferを返す
        ''' </summary>
        ''' <param name="obj">モデル</param>
        ''' <returns>変形されたVB</returns>
        Public Function TransMesh(ByVal obj As Model) As VertexBuffer

            Dim mes As Mesh = obj.Mesh.Clone(MeshFlags.Managed, VertexFormats.PositionNormal, Device)

            Dim v(mes.NumberVertices) As CustomVertex.PositionNormal
            Dim gs As GraphicsStream = mes.VertexBuffer.Lock(0, 0, LockFlags.None)

            v = CType(gs.Read(GetType(CustomVertex.PositionNormal), New Integer() {mes.NumberVertices}), CustomVertex.PositionNormal())

            Dim matr As Matrix = Matrix.Scaling(obj.Scaling)
            matr *= Matrix.RotationX(obj.Rotation.X)
            matr *= Matrix.RotationY(obj.Rotation.Y)
            matr *= Matrix.RotationZ(obj.Rotation.Z)

            For i As Integer = 0 To v.Length - 1
                v(i).Position = Vector3.TransformCoordinate(v(i).Position, matr)
            Next

            mes.VertexBuffer.Unlock()

            Dim returnbuf As VertexBuffer = New VertexBuffer(GetType(CustomVertex.PositionNormal), mes.NumberVertices, Device, Usage.None, CustomVertex.PositionNormal.Format, Pool.Managed)
            gs = returnbuf.Lock(0, 0, LockFlags.None)
            gs.Write(v)

            gs.Close()

            returnbuf.Unlock()

            Return returnbuf

        End Function

        ''' <summary>
        ''' 描画用フォントの新規登録
        ''' </summary>
        ''' <param name="FontName">フォント</param>
        ''' <param name="Size">フォントサイズ</param>
        ''' <returns>作成されたフォントのインデックス</returns>
        Public Function CreateFont(ByVal FontName As String, ByVal Size As Integer) As Integer

            Dim FontDe As FontDescription = New FontDescription
            With FontDe
                .FaceName = FontName
                .Height = Size
            End With

            Fonts.Add(New Direct3D.Font(Device, FontDe))

            Return Fonts.Count - 1

        End Function

        ''' <summary>
        ''' テクスチャの新規登録
        ''' </summary>
        ''' <param name="FileName">画像ファイル名</param>
        ''' <returns>作成されたテクスチャのインデックス</returns>
        Public Function CreateTexture(ByVal FileName As String) As Integer

            Dim Texture As Texture
            Texture = TextureLoader.FromFile(Device, FileName)

            Textures.Add(Texture)

            Return Textures.Count - 1

        End Function

        ''' <summary>
        ''' FPSの計測
        ''' </summary>
        Private Sub Timer_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs)
            FPSi = FPSTimeri
            FPSTimeri = 0
        End Sub


        Protected Overrides Sub OnFormClosing(ByVal e As System.Windows.Forms.FormClosingEventArgs)
            MyBase.OnFormClosing(e)
            If Not ClosingEvent Is Nothing Then ClosingEvent(e)
            Me.Dispose()

        End Sub

        Protected Overrides Sub OnKeyDown(ByVal e As System.Windows.Forms.KeyEventArgs)
            MyBase.OnKeyDown(e)
            If Not KeyEvent Is Nothing Then KeyEvent(e)

        End Sub

        Protected Overrides Sub OnMouseDown(ByVal e As System.Windows.Forms.MouseEventArgs)
            MyBase.OnMouseDown(e)
            If Not ClickEvent Is Nothing Then ClickEvent(e)
        End Sub

        Protected Overrides Sub Dispose(ByVal disposing As Boolean)
            MyBase.Dispose(disposing)

            'For Each x As Model In XModels
            '    x.Mesh.Dispose()
            '    For Each t As Texture In x.Texture
            '        If t Is Nothing Then Continue For
            '        t.Dispose()
            '    Next
            'Next
            Sprite.Dispose()
            Device.Dispose()

        End Sub


    End Class


    Public Class Input
        Implements IDisposable




        Dim InputDevice As DirectInput.Device

        Sub New(ByVal control As Control)

            For i As Integer = 0 To 20
                InputDevice = New DirectInput.Device(DirectInput.SystemGuid.Keyboard)

                If InputDevice Is Nothing Then Exit For
                Threading.Thread.Sleep(20)
            Next

            InputDevice.SetCooperativeLevel(control, CooperativeLevelFlags.Foreground Or CooperativeLevelFlags.NonExclusive)
        End Sub

        Function Get_Key(ByVal Key As DirectInput.Key) As Boolean
            Try
                InputDevice.Acquire()
            Catch ex As OtherApplicationHasPriorityException
                Return False
            End Try

            Dim keystate As KeyboardState = InputDevice.GetCurrentKeyboardState
            If keystate(Key) Then
                Return True
            End If

            Return False

        End Function

        Private disposedValue As Boolean = False        ' 重複する呼び出しを検出するには

        ' IDisposable
        Protected Overridable Sub Dispose(ByVal disposing As Boolean)
            If Not Me.disposedValue Then
                If disposing Then
                    ' TODO: 他の状態を解放します (マネージ オブジェクト)。
                End If
                InputDevice.Dispose()
                ' TODO: ユーザー独自の状態を解放します (アンマネージ オブジェクト)。
                ' TODO: 大きなフィールドを null に設定します。
            End If
            Me.disposedValue = True
        End Sub

#Region " IDisposable Support "
        ' このコードは、破棄可能なパターンを正しく実装できるように Visual Basic によって追加されました。
        Public Sub Dispose() Implements IDisposable.Dispose
            ' このコードを変更しないでください。クリーンアップ コードを上の Dispose(ByVal disposing As Boolean) に記述します。
            Dispose(True)
            GC.SuppressFinalize(Me)
        End Sub
#End Region

    End Class


    Public Class Sound
        Implements IDisposable

        Public Sounds As New List(Of Audio)

        Public Sub LoadAudio(ByVal Name As String, Optional ByVal Play As Boolean = False, Optional ByVal Volume As Integer = 0, Optional ByVal Balance As Integer = 0)

            Dim audio As Audio = New Audio(Name, Play)

            audio.Balance = Balance
            audio.Volume = Volume

            Sounds.Add(audio)
        End Sub

        Private disposedValue As Boolean = False        ' 重複する呼び出しを検出するには

        ' IDisposable
        Protected Overridable Sub Dispose(ByVal disposing As Boolean)
            If Not Me.disposedValue Then
                If disposing Then
                    ' TODO: 他の状態を解放します (マネージ オブジェクト)。
                End If
                For Each s As Audio In Sounds
                    s.Dispose()
                Next
                ' TODO: ユーザー独自の状態を解放します (アンマネージ オブジェクト)。
                ' TODO: 大きなフィールドを null に設定します。
            End If
            Me.disposedValue = True
        End Sub

#Region " IDisposable Support "
        ' このコードは、破棄可能なパターンを正しく実装できるように Visual Basic によって追加されました。
        Public Sub Dispose() Implements IDisposable.Dispose
            ' このコードを変更しないでください。クリーンアップ コードを上の Dispose(ByVal disposing As Boolean) に記述します。
            Dispose(True)
            GC.SuppressFinalize(Me)
        End Sub
#End Region

    End Class
End Namespace